@@ -19,7 +19,7 @@
 /* VTE accessibility object.  Based heavily on inspection of libzvt's
  * accessibility code. */
 
-#ident "$Id: vteaccess.c,v 1.25 2003/01/18 18:17:25 nalin Exp $"
+#ident "$Id: vteaccess.c,v 1.25.2.2 2003/02/11 00:50:26 nalin Exp $"
 #include "../config.h"
 #include <atk/atk.h>
 #include <gtk/gtk.h>
@@ -284,38 +284,42 @@ vte_terminal_accessible_update_private_data_if_needed(AtkObject *text,
 		/* We're finished updating this. */
 		priv->snapshot_contents_invalid = FALSE;
 	}
-	if (priv->snapshot_caret_invalid) {
-		vte_terminal_get_cursor_position(terminal, &ccol, &crow);
 
-		/* Get the offsets to the beginnings of each line. */
-		caret = -1;
-		for (i = 0; i < priv->snapshot_characters->len; i++) {
-			/* Get the attributes for the current cell. */
-			offset = g_array_index(priv->snapshot_characters,
-					       int, i);
-			attrs = g_array_index(priv->snapshot_attributes,
-					      struct vte_char_attributes,
-					      offset);
-			/* If this cell is "before" the cursor, move the
-			 * caret to be "here". */
-			if ((attrs.row < crow) ||
-			    ((attrs.row == crow) && (attrs.column < ccol))) {
-				caret = i + 1;
-			}
+	/* Update the caret position. */
+	vte_terminal_get_cursor_position(terminal, &ccol, &crow);
+
+	/* Get the offsets to the beginnings of each line. */
+	caret = -1;
+	for (i = 0; i < priv->snapshot_characters->len; i++) {
+		/* Get the attributes for the current cell. */
+		offset = g_array_index(priv->snapshot_characters,
+				       int, i);
+		attrs = g_array_index(priv->snapshot_attributes,
+				      struct vte_char_attributes,
+				      offset);
+		/* If this cell is "before" the cursor, move the
+		 * caret to be "here". */
+		if ((attrs.row < crow) ||
+		    ((attrs.row == crow) && (attrs.column < ccol))) {
+			caret = i + 1;
 		}
-		/* If no cells are before the caret, then the caret must be
-		 * at the end of the buffer. */
-		if (caret == -1) {
-			caret = priv->snapshot_characters->len;
-		}
-		/* The caret may have moved. */
-		if (caret != priv->snapshot_caret) {
-			priv->snapshot_caret = caret;
-			emit_text_caret_moved(G_OBJECT(text), caret);
-		}
-		/* Done updating the caret position, too. */
-		priv->snapshot_caret_invalid = FALSE;
 	}
+
+	/* If no cells are before the caret, then the caret must be
+	 * at the end of the buffer. */
+	if (caret == -1) {
+		caret = priv->snapshot_characters->len;
+	}
+
+	/* Notify observers if the caret moved. */
+	if (caret != priv->snapshot_caret) {
+		priv->snapshot_caret = caret;
+		emit_text_caret_moved(G_OBJECT(text), caret);
+	}
+
+	/* Done updating the caret position, whether we needed to or not. */
+	priv->snapshot_caret_invalid = FALSE;
+
 #ifdef VTE_DEBUG
 	if (_vte_debug_on(VTE_DEBUG_MISC)) {
 		fprintf(stderr, "Refreshed accessibility snapshot, "
@@ -592,6 +596,8 @@ vte_terminal_accessible_new(VteTerminal *terminal)
 	atk_object_initialize(ATK_OBJECT(access), G_OBJECT(terminal));
 
 	access->widget = GTK_WIDGET(terminal);
+	g_object_add_weak_pointer(G_OBJECT(terminal),
+				  (gpointer*)&access->widget);
 
 	g_object_set_data(G_OBJECT(access),
 			  VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA,
@@ -636,34 +642,46 @@ vte_terminal_accessible_finalize(GObject *object)
 	GtkAccessible *accessible = NULL;
         GObjectClass *gobject_class; 
 
+#ifdef VTE_DEBUG
+	if (_vte_debug_on(VTE_DEBUG_MISC)) {
+		fprintf(stderr, "Finalizing accessible peer.\n");
+	}
+#endif
+
 	g_return_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(object));
 	accessible = GTK_ACCESSIBLE(object);
 	gobject_class = g_type_class_peek_parent(VTE_TERMINAL_ACCESSIBLE_GET_CLASS(object));
 
-	g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
-					     G_SIGNAL_MATCH_FUNC |
-					     G_SIGNAL_MATCH_DATA,
-					     0, 0, NULL,
-					     vte_terminal_accessible_text_modified,
-					     object);
-	g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
-					     G_SIGNAL_MATCH_FUNC |
-					     G_SIGNAL_MATCH_DATA,
-					     0, 0, NULL,
-					     vte_terminal_accessible_text_scrolled,
-					     object);
-	g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
-					     G_SIGNAL_MATCH_FUNC |
-					     G_SIGNAL_MATCH_DATA,
-					     0, 0, NULL,
-					     vte_terminal_accessible_invalidate_cursor,
-					     object);
-	g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
-					     G_SIGNAL_MATCH_FUNC |
-					     G_SIGNAL_MATCH_DATA,
-					     0, 0, NULL,
-					     (gpointer)vte_terminal_accessible_title_changed,
-					     object);
+	if (accessible->widget != NULL) {
+		g_object_remove_weak_pointer(G_OBJECT(accessible->widget),
+					     (gpointer*) &accessible->widget);
+	}
+	if (G_IS_OBJECT(accessible->widget)) {
+		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
+						     G_SIGNAL_MATCH_FUNC |
+						     G_SIGNAL_MATCH_DATA,
+						     0, 0, NULL,
+						     vte_terminal_accessible_text_modified,
+						     object);
+		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
+						     G_SIGNAL_MATCH_FUNC |
+						     G_SIGNAL_MATCH_DATA,
+						     0, 0, NULL,
+						     vte_terminal_accessible_text_scrolled,
+						     object);
+		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
+						     G_SIGNAL_MATCH_FUNC |
+						     G_SIGNAL_MATCH_DATA,
+						     0, 0, NULL,
+						     vte_terminal_accessible_invalidate_cursor,
+						     object);
+		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
+						     G_SIGNAL_MATCH_FUNC |
+						     G_SIGNAL_MATCH_DATA,
+						     0, 0, NULL,
+						     (gpointer)vte_terminal_accessible_title_changed,
+						     object);
+	}
 	if (gobject_class->finalize != NULL) {
 		gobject_class->finalize(object);
 	}
@@ -751,6 +769,10 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 	}
 #endif
 	g_return_val_if_fail(priv->snapshot_text != NULL, g_strdup(""));
+	g_return_val_if_fail(priv->snapshot_characters != NULL, g_strdup(""));
+	if (offset == priv->snapshot_characters->len) {
+		return g_strdup("");
+	}
 	g_return_val_if_fail(offset < priv->snapshot_characters->len,
 			     g_strdup(""));
 	g_return_val_if_fail(offset >= 0, g_strdup(""));
@@ -1134,6 +1156,12 @@ vte_terminal_accessible_text_finalize(gpointer iface, gpointer data)
 	VteTerminalAccessiblePrivate *priv;
 	accessible_class = g_type_class_peek(GTK_TYPE_ACCESSIBLE); 
 
+#ifdef VTE_DEBUG
+	if (_vte_debug_on(VTE_DEBUG_MISC)) {
+		fprintf(stderr, "Finalizing accessible peer AtkText interface.\n");
+	}
+#endif
+
 	/* Free the private data. */
 	priv = g_object_get_data(G_OBJECT(iface),
 				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
